import re

from django.urls.resolvers import RoutePattern, _route_to_regex

from django.utils.translation import activate

from garpix_page.utils.get_garpix_page_models import get_garpix_page_models


class PageViewMixin:

    @classmethod
    def get_instance_by_slug(cls, slugs, languages_list):
        slug_list = slugs.split('/')

        lang = None

        if slug_list[0] in languages_list:
            lang = slug_list.pop(0)
            activate(lang)

        if len(slug_list) == 0:
            slug_list.append('')

        if len(slug_list) > 1 and slug_list[-1] == '':
            slug_list.pop(-1)

        url = f"/{'/'.join(slug_list)}"

        page_models = get_garpix_page_models()
        active_models = []
        for el in page_models:
            for key, value in el.url_patterns().items():
                pattern = RoutePattern(r'(<url>){}'.format(value['pattern']))
                pattern.regex = re.compile('^/(?P<url>.*){}$'.format(_route_to_regex(value['pattern'], pattern._is_endpoint)[0][1:]))
                match = pattern.match(url)
                if match is not None:
                    last, _, params = match

                    el_url = f"/{params.pop('url')}"

                    active_models.append({
                        'model': el,
                        'params': params,
                        'pattern': key,
                        'url': el_url,
                        'permissions': value.get('permissions', el.permissions)
                    })

        for model in active_models:
            instance = model['model'].active_on_site.filter(url=model['url']).first()

            if instance:
                instance.subpage_params = model['params']
                instance.subpage_key = model['pattern']
                instance.permissions = model['permissions']
                return instance

        return None



FILE_NAME_MODEL = '{app}/models/{page}_page.py'
FILE_NAME_MODEL_INIT = '{app}/models/__init__.py'

FILE_CONTENT_MODEL_BASE_PAGE = '''from django.db import models
from garpix_page.models import BasePage


class {page_capitalize}Page(BasePage):
    template = "pages/{page}.html"

    class Meta:
        verbose_name = "{page_capitalize}"
        verbose_name_plural = "{page_capitalize}s"
        ordering = ("-created_at",)
'''

FILE_CONTENT_MODEL_BASE_LIST_PAGE = '''from django.db import models
from garpix_page.models import BaseListPage


class {page_capitalize}Page(BaseListPage):
    paginate_by = 25
    template = 'pages/{page}.html'

    class Meta:
        verbose_name = "{page_capitalize}"
        verbose_name_plural = "{page_capitalize}s"
        ordering = ('-created_at',)
'''

FILE_CONTENT_MODEL_BASE_SEARCH_PAGE = '''from django.db import models
from garpix_page.models import BaseSearchPage


class {page_capitalize}Page(BaseSearchPage):
    paginate_by = 25
    template = 'pages/{page}.html'

    class Meta:
        verbose_name = "{page_capitalize}"
        verbose_name_plural = "{page_capitalize}s"
        ordering = ('-created_at',)
'''

FILE_CONTENT_MODEL_DICT = {
    'page': FILE_CONTENT_MODEL_BASE_PAGE,
    'list': FILE_CONTENT_MODEL_BASE_LIST_PAGE,
    'search': FILE_CONTENT_MODEL_BASE_SEARCH_PAGE,
}

FILE_CONTENT_MODEL_INIT = '''from .{page}_page import {page_capitalize}Page  # noqa
'''

FILE_NAME_TRANSLATION = '{app}/translation/{page}_page.py'
FILE_NAME_TRANSLATION_INIT = '{app}/translation/__init__.py'

FILE_CONTENT_TRANSLATION = '''from modeltranslation.translator import TranslationOptions, register
from ..models import {page_capitalize}Page


@register({page_capitalize}Page)
class {page_capitalize}PageTranslationOptions(TranslationOptions):
    pass
'''

FILE_CONTENT_TRANSLATION_INIT = '''from .{page}_page import {page_capitalize}PageTranslationOptions  # noqa
'''

FILE_NAME_ADMIN = '{app}/admin/{page}_page.py'
FILE_NAME_ADMIN_INIT = '{app}/admin/__init__.py'

FILE_CONTENT_ADMIN = '''from ..models.{page}_page import {page_capitalize}Page
from django.contrib import admin
from garpix_page.admin import BasePageAdmin


@admin.register({page_capitalize}Page)
class {page_capitalize}PageAdmin(BasePageAdmin):
    pass
'''

FILE_CONTENT_ADMIN_INIT = '''from .{page}_page import {page_capitalize}PageAdmin  # noqa
'''

FILE_NAME_TEMPLATE = '../frontend/templates/pages/{page}.html'

FILE_CONTENT_TEMPLATE_BASE_PAGE = '''{% extends 'base.html' %}

{% block content %}

{% include 'include/breadcrumb.html' %}

<h1>{{ object.title }}</h1>
{% endblock %}
'''

FILE_CONTENT_TEMPLATE_BASE_LIST_PAGE = '''{% extends 'base.html' %}

{% block content %}

{% include 'include/breadcrumb.html' %}

<h1>{{ object.title }}</h1>

{% for obj in paginated_object_list %}
    <div>
        <p><a href="{{ obj.get_absolute_url }}">{{ obj.title }}</a></p>
    </div>
{% empty %}
    <div>Записей пока нет.</div>
{% endfor %}

{% include 'include/pagination.html' %}

{% endblock %}
'''

FILE_CONTENT_TEMPLATE_BASE_SEARCH_PAGE = '''{% extends 'base.html' %}

{% block content %}

{% include 'include/breadcrumb.html' %}

<h1>{{ object.title }}</h1>

<form method="get">
  <div class="form-row align-items-center">
    <div class="col-auto">
      <label class="sr-only" for="q">Поиск</label>
      <input type="text" class="form-control mb-2" name="q" id="q" value="{{ q }}" placeholder="Введите поисковый запрос...">
    </div>
    <div class="col-auto">
      <button type="submit" class="btn btn-primary mb-2">Найти</button>
    </div>
  </div>
</form>

{% for obj in paginated_object_list %}
    <div>
        <p><a href="{{ obj.get_absolute_url }}">{{ obj.title }}</a></p>
    </div>
{% empty %}
    <div>Ничего не найдено.</div>
{% endfor %}

{% include 'include/pagination.html' %}

{% endblock %}
'''

FILE_CONTENT_TEMPLATE_DICT = {
    'page': FILE_CONTENT_TEMPLATE_BASE_PAGE,
    'list': FILE_CONTENT_TEMPLATE_BASE_LIST_PAGE,
    'search': FILE_CONTENT_TEMPLATE_BASE_SEARCH_PAGE,
}

FILE_NAME_APP_INIT = '{app}/__init__.py'
FILE_CONTENT_APP_INIT = '''default_app_config = '{app}.apps.{app_capitalize}Config'
'''

FILE_NAME_APPS = '{app}/apps.py'
FILE_CONTENT_APPS = '''from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class {app_capitalize}Config(AppConfig):
    name = '{app}'
    verbose_name = _('{app_capitalize}')
'''

FILE_NAME_MIGRATIONS_INIT = '{app}/migrations/__init__.py'



from garpix_page.settings import STRING_HANDLERS as s_h
from importlib import import_module as i_m
from functools import reduce

__all__ = ('apply_string_handling', )


obj_init = [getattr(i_m(s[:s.rfind('.')]), s[s.rfind('.') + 1:]) for s in set(s_h)]


def apply_string_handling(string, method='apply_tag_init'):
    return reduce(lambda s, C: getattr(C(), method)(s), obj_init, string)



from django.core.management.base import BaseCommand

from garpix_page.models import BasePage
from garpix_page.utils.set_children_urls import set_children_url


class Command(BaseCommand):
    help = 'Set old pages urls'

    def handle(self, *args, **options):
        BasePage.objects.rebuild()
        pages = BasePage.objects.filter(parent__isnull=True)

        pages_to_update = list(pages)

        for page in pages:
            page.set_url()

            children = page.get_children()

            set_children_url(page, children, pages_to_update)

        BasePage.objects.bulk_update(pages_to_update, ['url'])

        self.stdout.write(self.style.SUCCESS('Done'))




def set_children_url(instance, children, pages_to_update):
    pages_to_update += list(children)
    for page in children:
        page.parent = instance
        page.set_url()
        children = page.get_children()
        if any(children):
            set_children_url(page, children, pages_to_update)



from django.template.loader import render_to_string
from garpix_page.template import template_source
from garpix_page.utils import apply_string_handling

__all__ = ('get_render_html_value', )


def get_render_html_value(default_html, apply_django_tag=False):
    def _get_render_html_value():
        if not apply_django_tag:
            return apply_string_handling(
                template_source.get_template(default_html).template
            )

        return apply_string_handling(render_to_string(default_html))

    return _get_render_html_value



from garpix_page.models import GarpixPageSiteConfiguration
from django.contrib.sitemaps import GenericSitemap
from garpix_page.models.base_page import BasePage
from django.contrib.sitemaps.views import sitemap


def sitemap_view(request, section=None,
            template_name='sitemap.xml', content_type='application/xml'):

    pages = {
        'queryset': BasePage.on_site.filter(is_active=True, display_on_sitemap=True).order_by('title'),
        'date_field': 'created_at'
    }
    extra_data = {
        'priority': 1.0
    }
    try:
        config = GarpixPageSiteConfiguration.get_solo()
        extra_data.update({
            'changefreq': config.sitemap_frequency
        })
    except Exception:
        pass

    sitemaps = {
            'pages': GenericSitemap(
                pages,
                **extra_data
            ),
    }
    return sitemap(request, sitemaps, section, template_name, content_type)



from .base_component import BaseComponent
from garpix_page.fields import GrapesJsHtmlField
from django.conf import settings

MEDIA_URL = settings.MEDIA_URL


class GrapesJsHtmlComponent(BaseComponent):
    html = GrapesJsHtmlField()
    admin_preview_image = 'garpix_page/images/admin/grapesjs.png'

    def get_context(self, request):
        context = super().get_context(request)
        context.update({
            'css': request.build_absolute_uri(f'{MEDIA_URL}grapesjs/html_{self.pk}_css.css'),
            'js': request.build_absolute_uri(f'{MEDIA_URL}grapesjs/html_{self.pk}_js.js')
        })
        return context

    class Meta:
        verbose_name = 'GrapesJs компонент'
        verbose_name_plural = 'GrapesJs компоненты'



from django import forms
from garpix_page.settings import GRAPESJS_TEMPLATE
from garpix_page.utils import get_render_html_value

__all__ = (
    'GrapesJsWidget',
)


class GrapesJsWidget(forms.Textarea):
    '''
    Textarea form widget with support grapesjs.
    This is widget base config grapesjs.
    '''
    template_name = GRAPESJS_TEMPLATE

    class Media:
        css = {
            'all': (
                'garpix_page/css/grapesjs/grapes.min.css',
                'garpix_page/css/grapesjs/grapesjs-preset-newsletter.css',
                'garpix_page/css/grapesjs/grapesjs-preset-webpage.min.css',
                'garpix_page/css/grapesjs/grapesjs-plugin-filestack.css',
            )
        }
        js = [
            'garpix_page/js/grapesjs/grapes.js',
            'garpix_page/js/grapesjs/grapesjs-preset-webpage.js',
            'garpix_page/js/grapesjs/grapesjs-blocks-basic.js',
            'garpix_page/js/grapesjs/grapesjs-plugin-forms.js',
            'garpix_page/js/grapesjs/grapesjs-component-countdown.js',
            'garpix_page/js/grapesjs/grapesjs-plugin-export.js',
            'garpix_page/js/grapesjs/grapesjs-tabs.js',
            'garpix_page/js/grapesjs/grapesjs-custom-code.js',
            'garpix_page/js/grapesjs/grapesjs-touch.js',
            'garpix_page/js/grapesjs/grapesjs-parser-postcss.js',
            'garpix_page/js/grapesjs/grapesjs-tooltip.js',
            'garpix_page/js/grapesjs/grapesjs-tui-image-editor.js',
            'garpix_page/js/grapesjs/grapesjs-typed.js',
            'garpix_page/js/grapesjs/grapesjs-style-bg.js',
            'garpix_page/js/grapesjs/ckeditor.js',
            'garpix_page/js/grapesjs/grapesjs-plugin-ckeditor.js',
            'garpix_page/js/grapesjs/grapesjs-preset-newsletter.js',  # custom build
        ]

    def get_formatted_id_value(self, value_id):
        return value_id.replace('-', '_')

    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)

        context['widget']['attrs']['id'] = self.get_formatted_id_value(context['widget']['attrs']['id'])
        context['widget'].update({
            'get_render_html_value': get_render_html_value(
                self.default_html, apply_django_tag=self.apply_django_tag
            ),
            'html_name_init_conf': self.html_name_init_conf,
            'template_choices': self.template_choices,
            'apply_django_tag': int(self.apply_django_tag),
        })

        return context



from django.utils.html import format_html
from modeltranslation.admin import TabbedTranslationAdmin

from ..forms import SeoTemplateForm
from garpix_page.models import SeoTemplate

from django.contrib import admin


@admin.register(SeoTemplate)
class SeoTemplateAdmin(TabbedTranslationAdmin, admin.ModelAdmin):

    form = SeoTemplateForm

    list_display = ['__str__', 'priority_order']

    readonly_fields = ('get_seo_template_keys', )

    def get_seo_template_keys(self, obj):
        from garpix_page.utils.get_garpix_page_models import get_garpix_page_models
        seo_template_keys_field = ''
        seo_template_keys = []
        for page_model in get_garpix_page_models():
            for field in page_model.seo_template_keys_list():
                if field[0] not in seo_template_keys:
                    seo_template_keys.append(field[0])
                    seo_template_keys_field += f'<b>{field[0]}</b> ({field[1]})</br>'
        return format_html(seo_template_keys_field)

    get_seo_template_keys.short_description = 'Доступные ключи'



from django.contrib import admin

from garpix_utils.admin import GarpixSiteConfigurationAdmin
from garpix_page.models.settings import GarpixPageSiteConfiguration


@admin.register(GarpixPageSiteConfiguration)
class GarpixPageSiteConfigurationAdmin(GarpixSiteConfigurationAdmin):
    pass



from django.db import models
from django.urls import reverse
from django.utils.html import format_html
from garpix_utils.managers import AvailableManager
from polymorphic.managers import PolymorphicManager

from ...mixins import CloneMixin
from ...models import BasePage
from polymorphic.models import PolymorphicModel
from ...serializers import get_components_serializer


class PageComponent(models.Model):
    component = models.ForeignKey("BaseComponent", on_delete=models.CASCADE, verbose_name='Компонент')
    page = models.ForeignKey("BasePage", on_delete=models.CASCADE, verbose_name='Страница')
    view_order = models.IntegerField(default=1, verbose_name='Порядок отображения')

    def __str__(self):
        return ''

    class Meta:
        unique_together = (('component', 'page'),)
        ordering = ('view_order', )
        verbose_name = 'Компонент страницы'
        verbose_name_plural = 'Компоненты страницы'


class BaseComponent(CloneMixin, PolymorphicModel):
    """
    Базовый компонент
    """
    title = models.CharField(max_length=255, verbose_name='Название')
    is_deleted = models.BooleanField(default=False, verbose_name='Запись удалена')
    is_active = models.BooleanField(default=True, verbose_name='Включено')
    created_at = models.DateTimeField(auto_now_add=True, verbose_name='Дата создания')
    updated_at = models.DateTimeField(auto_now=True, verbose_name='Дата изменения')

    pages = models.ManyToManyField(BasePage, blank=True, related_name='components', through='PageComponent',
                                   verbose_name='Страницы для отображения')

    text_title = models.CharField(blank=True, default='', max_length=128, verbose_name='Заголовок')
    template = 'garpix_page/components/default.html'

    searchable_fields = ('title',)
    serializer = None
    objects = PolymorphicManager()
    active_objects = AvailableManager()

    class Meta:
        verbose_name = 'Компонент'
        verbose_name_plural = 'Компоненты'
        ordering = ('created_at', 'title',)

    def __str__(self):
        return self.title

    def get_context(self, request):
        context = {
            'object': self,
            "component_model": self.__class__.__name__,
            'admin_edit_url': self.get_admin_url_edit_object()
        }
        return context

    def get_template(self):
        return self.template

    def model_name(self):
        real_instance = self.get_real_instance_class()
        if real_instance:
            return real_instance._meta.verbose_name  # noqa
        return '- нет -'

    model_name.short_description = 'Тип'

    @classmethod
    def is_for_component_view(cls):
        return True

    @property
    def admin_link_to_change(self):
        link = reverse("admin:garpix_page_basecomponent_change",
                       args=[self.id])
        return format_html('<a class="inlinechangelink" href="{0}">{1}</a>', link, self.title)

    def get_context_data(self, request):
        context = self.get_context(request)

        context.update({
            'template': self.get_template()
        })
        return context

    def get_api_context_data(self, request):
        context = self.get_context(request)
        for k, v in context.items():
            if hasattr(v, 'is_for_component_view'):
                model_serializer_class = get_components_serializer(v.__class__)
                context[k] = model_serializer_class(v, context={"request": request}).data
        return context

    def get_serializer(self):
        return None

    def get_admin_url_edit_object(self):
        url = reverse(f'admin:{self._meta.app_label}_{self._meta.model_name}_change', args=[self.id])
        return url

    def delete(self, using=None, keep_parents=False):
        self.is_deleted = True
        self.save()

    def hard_delete(self):
        super().delete()



from django.conf import settings
from django.core.cache import cache

from garpix_page.utils.all_sites import get_all_sites


class PageCacheService:
    cache_url_prefix = 'url_page_'
    cache_instance_prefix = 'instance_page_'

    def get_url(self, pk, current_language_code_url_prefix):
        current_site = getattr(settings, 'SITE_ID', 1)
        cache_key = f'{self.cache_url_prefix}{pk}_{current_site}_{current_language_code_url_prefix}'
        url_cache = cache.get(cache_key)
        if url_cache is not None:
            return url_cache
        return None

    def set_url(self, pk, current_language_code_url_prefix, result):
        current_site = getattr(settings, 'SITE_ID', 1)
        cache_key = f'{self.cache_url_prefix}{pk}_{current_site}_{current_language_code_url_prefix}'
        cache.set(cache_key, result)

    def get_instance_by_url(self, url):
        current_site = getattr(settings, 'SITE_ID', 1)
        cache_key = f'{self.cache_instance_prefix}_{current_site}_{url}'
        url_cache = cache.get(cache_key)
        if url_cache is not None:
            return url_cache
        return None

    def set_instance_by_url(self, url, result):
        current_site = getattr(settings, 'SITE_ID', 1)
        cache_key = f'{self.cache_instance_prefix}_{current_site}_{url}'
        cache.set(cache_key, result)

    def set_seo_by_page(self, pk, field_name, result, site):
        cache_key = f'page_{field_name}_{site}_{pk}'
        cache.set(cache_key, result)

    def get_seo_by_page(self, pk, field_name, site):
        cache_key = f'page_{field_name}_{site}_{pk}'
        seo_cache = cache.get(cache_key)
        if seo_cache is not None:
            return seo_cache
        return None

    def clear_seo_data(self, pk=None):
        from garpix_page.models import BasePage
        seo_fields = [field_name for field in BasePage._meta.get_fields() if (field_name := field.name)[:4] == 'seo_']
        keys = []
        if pk:
            for seo_field in seo_fields:
                for site in get_all_sites():
                    keys.append(f'page_{seo_field}_{site}_{pk}')
        else:
            for page in BasePage.objects.all():
                for seo_field in seo_fields:
                    for site in get_all_sites():
                        keys.append(f'page_{seo_field}_{site}_{page.pk}')

        cache.delete_many(keys=keys)

    def clear_all_by_page(self, instance, current_language_code_url_prefix):
        url, pk = instance.get_absolute_url(), instance.pk
        cache.delete(f'{self.cache_url_prefix}{pk}_{current_language_code_url_prefix}')
        cache.delete(f'{self.cache_instance_prefix}{url}')
        self.clear_seo_data(pk)

    def reset_url_info_by_page(self, instance, current_language_code_url_prefix):
        pk, old_url = instance.pk, instance.get_absolute_url()
        current_site = getattr(settings, 'SITE_ID', 1)
        url_cache_key = f'{self.cache_url_prefix}{pk}_{current_site}_{current_language_code_url_prefix}'
        instance_cache_key = f'{self.cache_instance_prefix}_{current_site}_{old_url}'
        cache.delete(url_cache_key)
        cache.delete(instance_cache_key)

        new_url = instance.get_absolute_url()
        self.set_instance_by_url(new_url, instance)

    def clear_all(self):
        cache.clear()


cache_service = PageCacheService()



from .base_page import BasePage
from garpix_utils.paginator import GarpixPaginator

from ..pagination import GarpixPagePagination
from ..serializers import get_serializer


class BaseListPage(BasePage):
    paginate_by = 25
    template = 'garpix_page/default_list.html'

    def get_context(self, request=None, *args, **kwargs):
        context = super().get_context(request, *args, **kwargs)

        object_list = self.children.filter(is_active=True)
        paginator = GarpixPaginator(object_list, self.paginate_by)

        try:
            page = int(request.GET.get('page', 1))
        except ValueError:
            page = 1

        paginated_object_list = paginator.get_page(page)

        api_paginator = GarpixPagePagination()

        try:
            model_serializer_class = get_serializer(object_list[0].__class__)

            paginated_children_list = api_paginator.get_paginated_data(queryset=object_list,
                                                                       serializer=model_serializer_class,
                                                                       request=request)
        except Exception:
            paginated_children_list = {
                "count": 0,
                "next": None,
                "prev": None,
                "results": []
            }

        context.update({
            'paginator': paginator,
            'paginated_object_list': paginated_object_list,
            'page': page,
            'paginated_children_list': paginated_children_list
        })
        return context

    class Meta:
        abstract = True



from django.conf import settings
from django.utils.module_loading import import_string

from garpix_page.utils.set_children_urls import set_children_url

celery_app = import_string(settings.GARPIXCMS_CELERY_SETTINGS)


@celery_app.task()
def clear_child_cache(id):
    from garpix_page.models import BasePage

    instance = BasePage.objects.get(pk=id)
    children = instance.get_children()

    pages_to_update = []
    set_children_url(instance, children, pages_to_update)

    BasePage.objects.bulk_update(pages_to_update, ['url'])
    BasePage.objects.rebuild()



import re
from functools import reduce
from django.template import Template, Context
from garpix_page.settings import NAME_MAKEUP_TAG

__all__ = ('ApplyMakeupTag', )


REGEX_MAKEUP_TAG = '<%s>(.*?)</%s>' % (NAME_MAKEUP_TAG, NAME_MAKEUP_TAG)
REGEX_MAKEUP_INIT = '<%s>(?!<div hidden="">{#).*?(?!#}</div>)</%s>' % (NAME_MAKEUP_TAG, NAME_MAKEUP_TAG)
REGEX_MAKEUP_SAVE = '<%s.*?><div hidden="">{#(.*?)#}</div>.*?</%s>' % (NAME_MAKEUP_TAG, NAME_MAKEUP_TAG)


class ApplyMakeupTag(object):
    def apply_tag_init(self, string):
        strings_to_render = re.findall(REGEX_MAKEUP_TAG, string)

        replace_to_strings = map(
            lambda t:
                '<%s><div hidden="">{#%s#}</div>%s</%s>' % (
                    NAME_MAKEUP_TAG, t.source, t.render(Context({})), NAME_MAKEUP_TAG
                ), map(Template, strings_to_render)
        )

        return reduce(lambda s, r: re.sub(REGEX_MAKEUP_INIT, r, s, 1), replace_to_strings, string)

    def apply_tag_save(self, string):
        substrings_to_save = [
            '<%s>%s</%s>' % (NAME_MAKEUP_TAG, sub, NAME_MAKEUP_TAG) for sub in re.findall(REGEX_MAKEUP_SAVE, string)
        ]

        return reduce(lambda s, r: re.sub(REGEX_MAKEUP_SAVE, r, s, 1), substrings_to_save, string)



from django import forms
from django.conf import settings
from django.core.exceptions import ValidationError
from django.utils.functional import cached_property
from polymorphic.admin import PolymorphicModelChoiceForm
from django.contrib.contenttypes.models import ContentType
from django.apps import apps
from polymorphic_tree.admin import PolymorpicMPTTAdminForm
from django.utils.translation import gettext as _
from garpix_page.models import BasePage

from django.db.models import CharField, TextField
from garpix_page.utils.get_garpix_page_models import get_garpix_page_models
from garpix_page.utils.get_languages import get_languages


class AdminRadioSelectPreview(forms.RadioSelect):
    template_name = 'garpix_page/admin/widgets/radio_preview.html'

    def create_option(self, name, value, *args, **kwargs):
        result = super().create_option(name, value, *args, **kwargs)
        ct = ContentType.objects.filter(pk=value).first()
        result['attrs']['preview'] = []
        result['attrs']['group'] = None
        if ct is not None:
            model = apps.get_model(ct.app_label, ct.model)
            preview = getattr(model, 'admin_preview_image', [])
            group = getattr(model, 'admin_group', None)
            result['attrs']['preview'] = preview if isinstance(preview, list) else [preview]
            result['attrs']['group'] = group
        return result

    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)
        context["widget"]["sorted_groups"] = self.sorted_groups(
            context["widget"]["optgroups"]
        )
        return context

    def sorted_groups(self, optgroups):
        groups = optgroups
        result = {}
        for group in groups:
            subgroups = group[1]
            for subgroup in subgroups:
                groupKey = subgroup['attrs']['group']
                if groupKey not in result:
                    result[groupKey] = []
                result[groupKey].append(subgroup)
        return result


class PolymorphicModelPreviewChoiceForm(PolymorphicModelChoiceForm):
    class Media:
        css = {
            'all': ('garpix_page/css/admin/styles.css',)
        }
        js = ('garpix_page/js/admin/components.js',)

    type_label = 'Тип'

    ct_id = forms.ChoiceField(
        label=type_label, widget=AdminRadioSelectPreview(attrs={"class": "radiolist"})
    )

    def __init__(self, *args, **kwargs):
        # Allow to easily redefine the label (a commonly expected usecase)
        super().__init__(*args, **kwargs)
        self.fields["ct_id"].label = self.type_label


class PageForm(PolymorpicMPTTAdminForm):
    def clean(self):
        cleaned_data = super().clean()

        sites = cleaned_data.get('sites')
        slug = cleaned_data.get('slug')
        parent = cleaned_data.get('parent')

        self.instance.slug = slug
        self.instance.set_url(parent)

        languages = [x[0] for x in settings.LANGUAGES]

        pages = BasePage.objects.filter(url=self.instance.url).exclude(pk=self.instance.pk)

        for page in pages:
            if page.absolute_url == self.instance.url and any(set(sites) & set(page.sites.all())):
                raise ValidationError({'slug': 'Страница с таким ЧПУ существует'})
        if slug in languages:
            raise ValidationError({'slug': f'ЧПУ не должен совпадать с языковым кодом ({languages})'})

        return cleaned_data


def get_model_rule_fields_values():
    model_rule_fields_values = []
    for page_model in get_garpix_page_models():
        model_rule_fields_values.append((page_model.__name__, page_model._meta.verbose_name))
    return model_rule_fields_values


def get_rule_fields_values():
    rule_fields_values = []
    for page_model in get_garpix_page_models():
        for f in page_model._meta.get_fields():
            if hasattr(f, 'verbose_name') and (
                    isinstance(f, CharField) or isinstance(f, TextField)) and 'seo_' not in f.name and f.name != 'slug':
                prop_field = (f.name, f.verbose_name)
                if prop_field not in rule_fields_values:
                    rule_fields_values.append(prop_field)

        for f in [f for name in dir(page_model) if isinstance(f := getattr(page_model, name), cached_property)]:
            prop_field = (f.name, f.short_description)
            if prop_field not in rule_fields_values:
                rule_fields_values.append(prop_field)
    return rule_fields_values


class SeoTemplateForm(forms.ModelForm):
    class RULE_FIELD:
        MODEL_NAME = 'model_name'

        CHOICES = (
                      (MODEL_NAME, 'Название модели'),
                  ) + tuple(get_rule_fields_values())

    rule_field = forms.ChoiceField(choices=RULE_FIELD.CHOICES)
    rule_field.label = 'Поле'
    model_rule_value = forms.ChoiceField(choices=tuple(get_model_rule_fields_values()))
    model_rule_value.label = 'Название'

    class Media:
        css = {
            'all': ('garpix_page/css/admin/seo_template.css',)
        }
        js = ('garpix_page/js/admin/seo_template.js',)

    def clean(self):
        _data = self.cleaned_data
        lang_seo_fields = ['seo_title', 'seo_keywords', 'seo_description', 'seo_author']
        seo_fields = ['seo_og_type']

        langs = get_languages()
        for field_name in lang_seo_fields:
            for lang in langs:
                lang = lang.replace('-', '_')
                seo_fields.append(f"{field_name}_{lang}")

        models_fields_values = {field[0]: 'value' for field in get_rule_fields_values()}

        for field_name in seo_fields:
            try:
                seo_value = str(_data.get(field_name, '')).format(
                    **models_fields_values)
            except (AttributeError, KeyError, ValueError) as e:
                raise ValidationError({field_name: _("Некорректный шаблон")})



from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class GarpixPageConfig(AppConfig):
    name = 'garpix_page'
    verbose_name = _('Pages')

    def ready(self):
        import garpix_page.signals  # noqa



from django.views.generic import View
from django.views.decorators.csrf import csrf_exempt
from django.utils.decorators import method_decorator
from django.http import JsonResponse
from django.core.files.storage import FileSystemStorage


@method_decorator(csrf_exempt, name='dispatch')
class DgjsUpload(View):
    def post(self, request, *args, **kwargs):
        user = request.user
        if user.is_staff:
            files = request.FILES.getlist('files[]')
            data = []
            fs = FileSystemStorage()
            for file in files:
                filename = fs.save(file.name, file)
                file_url = fs.url(filename)
                data.append(request.build_absolute_uri(file_url))
            return JsonResponse({'data': data})
        return JsonResponse({'data': []})



from django import template

register = template.Library()


@register.simple_tag(takes_context=True)
def gx_component(context, component):
    user = context['user']
    if not user.is_authenticated or not user.is_staff:
        return ''

    component_path = None
    if isinstance(component, dict):
        component_path = component.get('admin_edit_url')
    elif hasattr(component, 'admin_edit_url'):
        component_path = getattr(component, 'admin_edit_url')

    if component_path:
        return f'data-gx-component={component_path}'
    else:
        return ''



from django.conf import settings
from django.db.models.signals import pre_save, pre_delete
from django.dispatch import receiver
from django.utils import translation
from django.utils.functional import cached_property
from django.db import models
from django.urls import reverse
from django.contrib.sites.models import Site
from rest_framework.views import APIView
from garpix_page.utils.all_sites import get_all_sites
from garpix_page.utils.get_file_path import get_file_path
from polymorphic_tree.models import PolymorphicMPTTModel, PolymorphicTreeForeignKey, PolymorphicMPTTModelManager
from django.utils.html import format_html
from garpix_utils.managers import GCurrentSiteManager, GPolymorphicCurrentSiteManager, ActiveOnSiteManager
from ..cache import cache_service
from ..mixins import CloneMixin
from garpix_admin_lock.mixins import PageLockViewMixin

from ..tasks import clear_child_cache
from ..utils.get_current_language_code_url_prefix import get_current_language_code_url_prefix
from ..utils.set_children_urls import set_children_url


class BasePage(CloneMixin, PolymorphicMPTTModel, PageLockViewMixin):
    """
    Базовая страница, на основе которой создаются все прочие страницы.
    """
    title = models.CharField(max_length=255, verbose_name='Название')
    is_active = models.BooleanField(default=True, verbose_name='Включено')
    display_on_sitemap = models.BooleanField(default=True, verbose_name='Отображать в карте сайта')
    slug = models.SlugField(max_length=150, verbose_name='ЧПУ', blank=True, default='')
    sites = models.ManyToManyField(Site, default=get_all_sites, verbose_name='Сайты для отображения')
    created_at = models.DateTimeField(auto_now_add=True, verbose_name='Дата создания')
    updated_at = models.DateTimeField(auto_now=True, verbose_name='Дата изменения')
    seo_title = models.CharField(max_length=250, verbose_name='SEO заголовок страницы (title)', blank=True, default='')
    seo_keywords = models.CharField(max_length=250, verbose_name='SEO ключевые слова (keywords)', blank=True,
                                    default='')
    seo_description = models.TextField(verbose_name='SEO описание (description)', blank=True, default='')
    seo_author = models.CharField(max_length=250, verbose_name='SEO автор (author)', blank=True, default='')
    seo_og_type = models.CharField(max_length=250, verbose_name='SEO og:type', blank=True, default="website")
    seo_image = models.FileField(upload_to=get_file_path, blank=True, null=True, verbose_name='SEO изображение')
    parent = PolymorphicTreeForeignKey('self', null=True, blank=True, related_name='children',
                                       db_index=True, verbose_name='Родительская страница', on_delete=models.SET_NULL,
                                       limit_choices_to={})
    url = models.CharField(max_length=255, default='', blank=True, verbose_name='Полный URL страницы')

    # objects = models.Manager()
    objects = PolymorphicMPTTModelManager()
    on_site = GCurrentSiteManager()
    polymorphic_on_site = GPolymorphicCurrentSiteManager()
    active_on_site = ActiveOnSiteManager()

    template = 'garpix_page/default.html'
    searchable_fields = ('title',)
    serializer = None  # default is generator of serializers: garpix_page.serializers.serializer.get_serializer
    permissions = None
    subpage_params = None
    subpage_key = None

    class Meta(PolymorphicMPTTModel.Meta):
        verbose_name = 'Структура страниц'
        verbose_name_plural = 'Структура страниц'
        ordering = ('created_at', 'title',)

    def get_model_class_name(self):
        if self.subpage_key:
            return str(self.subpage_key).format(model_name=self.__class__.__name__)
        return self.__class__.__name__

    def __str__(self):
        return self.title

    @cached_property
    def _default_site(self):
        return Site.objects.get(pk=getattr(settings, 'SITE_ID', 1))

    _default_site.short_description = 'Default Site'

    def get_seo_template_keys(self):
        return self.__dict__

    @classmethod
    def seo_template_keys_list(cls):
        return [(f.name, f.verbose_name) for f in cls._meta.fields if hasattr(f, 'verbose_name') and (
                isinstance(f, models.CharField) or isinstance(f, models.TextField))]

    def get_verbose_model_name(self):
        return self._meta.verbose_name

    get_verbose_model_name.short_description = 'Название модели'

    def get_absolute_url(self):
        return self.absolute_url

    get_absolute_url.short_description = 'URL'

    def get_absolute_url_html(self):
        return f'<a href="{self.absolute_url}" target="_blank">{self.absolute_url}</a>'

    get_absolute_url_html.short_description = 'URL'
    get_absolute_url_html.allow_tags = True

    @cached_property
    def absolute_url(self):
        current_language_code_url_prefix = get_current_language_code_url_prefix()
        return "{}{}".format(current_language_code_url_prefix,
                             self.url) if current_language_code_url_prefix else self.url

    absolute_url.short_description = 'URL'

    def set_url(self, parent=None):

        parent = parent or self.parent

        self.url = ''

        if parent and parent.url != '/':
            self.url = parent.url

        self.url += f"/{self.slug}"

    @cached_property
    def get_sites(self):
        res = 'n/a'
        if self.sites.all().count() > 0:
            res = ''
            for site in self.sites.all():
                res += f'{site.domain} '
        return res

    get_sites.short_description = 'Sites'

    def get_template(self):
        return self.template

    def get_context(self, request=None, *args, **kwargs):

        return {
            'object': self,
            'subpages': self.get_subpages_list(),
            'components': self.get_components_context(request, api=kwargs.get('api', False))
        }

    @classmethod
    def is_for_page_view(cls):
        return True

    def get_breadcrumbs(self):
        result = []
        obj = self
        result.append(obj)
        while obj.parent is not None:
            result.insert(0, obj.parent)
            obj = obj.parent
        return result

    def get_admin_url_edit_object(self):
        url = reverse(f'admin:{self._meta.app_label}_{self._meta.model_name}_change', args=[self.id])
        return url

    def get_serializer(self):
        return self.serializer

    def model_name(self):
        real_instance = self.get_real_instance_class()
        if real_instance:
            return real_instance._meta.verbose_name  # noqa
        return '- нет -'

    model_name.short_description = 'Тип'

    def has_permission_required(self, request):

        view = APIView()
        view.queryset = type(self).objects.filter(id=self.id)
        if self.permissions is not None:
            for permission in [permission() for permission in self.permissions]:
                if not permission.has_object_permission(request, view, self) or not permission.has_permission(request,
                                                                                                              view):
                    return False
        return True

    def get_components_context(self, request, api=False):
        context = []
        components = self.pagecomponent_set.filter(component__is_active=True, component__is_deleted=False).order_by(
            'view_order')
        for component in components:
            component_context = {
                'view_order': component.view_order
            }
            if api:
                component_context.update(component.component.get_api_context_data(request))
            else:
                component_context.update(component.component.get_context_data(request))
            context.append(component_context)
        return context

    def get_components(self):
        context = []
        components = self.pagecomponent_set.filter(component__is_active=True, component__is_deleted=False)
        for component in components:
            context.append(component.component)
        return context

    def admin_link_to_add_component(self):
        link = reverse("admin:garpix_page_basecomponent_add")
        return format_html(
            '<a class="related-widget-wrapper-link add-related addlink" href="{0}?_to_field=id&_popup=1&pages={1}">Добавить компонент</a>',
            link, self.id)

    def get_seo_value(self, field_name, site=None):
        from garpix_page.admin.settings.seo_template import SeoTemplateForm
        from garpix_page.models.settings import SeoTemplate

        site = site or self._default_site

        seo_value_cache = cache_service.get_seo_by_page(self.pk, field_name, site)

        if seo_value_cache is not None:
            return seo_value_cache

        seo_templates = SeoTemplate.active_objects.filter(sites__in=[site]).all()

        for temp in seo_templates:
            is_model_rule = temp.rule_field == SeoTemplateForm.RULE_FIELD.MODEL_NAME and self.__class__.__name__ == temp.model_rule_value
            is_field_rule = temp.rule_field != SeoTemplateForm.RULE_FIELD.MODEL_NAME and str(temp.rule_value) in str(
                getattr(self, temp.rule_field, None))
            if is_model_rule or is_field_rule:
                try:
                    seo_value = getattr(temp, field_name, '').format(**self.get_seo_template_keys())
                except (AttributeError, KeyError, ValueError) as e:
                    # ToDo: добавить предупреждение в админке
                    print(f'{field_name}: {e}')
                    seo_value = getattr(temp, field_name, '')
                break
        else:
            seo_value = getattr(self, field_name, '')

        cache_service.set_seo_by_page(self.pk, field_name, seo_value, site)
        return seo_value or ''

    def get_seo_title(self):
        language_code = translation.get_language()
        return self.get_seo_value(f'seo_title_{language_code}')

    def get_seo_keywords(self):
        language_code = translation.get_language()
        return self.get_seo_value(f'seo_keywords_{language_code}')

    def get_seo_description(self):
        language_code = translation.get_language()
        return self.get_seo_value(f'seo_description_{language_code}')

    def get_seo_author(self):
        language_code = translation.get_language()
        return self.get_seo_value(f'seo_author_{language_code}')

    def get_seo_og_type(self):
        return self.get_seo_value('seo_og_type')

    def get_seo_image(self):
        return self.get_seo_value('seo_image')

    @classmethod
    def url_patterns(cls):
        return {
            '{model_name}': {
                'verbose_name': '{model_title}',
                'pattern': ''
            },
        }

    def get_subpages_list(self):
        patterns = self.url_patterns()
        patterns.pop('{model_name}')
        subpages = {}
        for key, pattern in patterns.items():
            subpages[key] = {'title': pattern['verbose_name'],
                             'absolute_url': f"{self.absolute_url}{pattern['pattern']}"}
        return subpages

    def get_subpage_url(self, subpage_key=None, subpage_params=None):
        subpage_key = subpage_key or self.subpage_key
        if subpage_key:
            parameters = subpage_params or self.subpage_params
            sub_page_url = self.url_patterns()[subpage_key]['pattern'].replace('<', '%(').replace('>',
                                                                                                  ')s') % parameters
            return f"{self.get_absolute_url()}{sub_page_url}"
        return ''


@receiver(pre_save)
def reset_url(sender, instance: BasePage, update_fields, **kwargs):
    if type(sender) == type(BasePage):
        if instance.seo_title is None:
            instance.seo_title = instance.title

        if instance.pk:

            cache_service.clear_seo_data(instance.pk)
            old_instance = BasePage.objects.get(pk=instance.pk)
            if instance.parent != old_instance.parent or instance.slug != old_instance.slug:

                instance.set_url()
                children = instance.get_children()
                if children:
                    if len(children) > getattr(settings, 'GARPIX_PAGE_CHILDREN_LEN', 10):
                        clear_child_cache.delay(instance.id)
                    else:
                        pages_to_update = []
                        set_children_url(instance, children, pages_to_update)

                        BasePage.objects.bulk_update(pages_to_update, ['url'])
                        BasePage.objects.rebuild()

        else:
            instance.set_url()


@receiver(pre_delete)
def update_children_url(sender, instance: BasePage, *args, **kwargs):
    if type(sender) == type(BasePage):
        children = instance.get_children()
        if children:
            pages_to_update = []
            set_children_url(None, children, pages_to_update)

            BasePage.objects.bulk_update(pages_to_update, ['url'])
            BasePage.objects.rebuild()



from modeltranslation.translator import TranslationOptions, register
from garpix_page.models import GrapesJsHtmlComponent


@register(GrapesJsHtmlComponent)
class GrapesJsHtmlComponentTranslationOptions(TranslationOptions):
    pass
